@@ -157,13 +157,24 @@ typedef struct {
 static int
 initialize_globals(void)
 {
-    static int initialized = 0;
+    static int initialized = -1;
     PyObject *module;
-    
-    if (initialized)
-        return 1;
 
-    initialized = 1;
+    if (initialized >= 0) {
+        /* This function should never fail under normal circumstances,
+         * but if it does, raise an error on subsequent calls instead
+         * of segfaulting.  This should make it easier to track down
+         * the cause of such errors.
+         *
+         * https://bugs.launchpad.net/storm/+bug/1006284
+         */
+        if (!initialized)
+            PyErr_SetString(PyExc_RuntimeError,
+                            "initialize_globals() failed the first time "
+                            "it was run");
+        return initialized;
+    }
+    initialized = 0;
 
     /* Import objects from storm module */
     module = PyImport_ImportModule("storm");
@@ -195,7 +206,7 @@ initialize_globals(void)
     module = PyImport_ImportModule("storm.info");
     if (!module)
         return 0;
-    
+
     get_cls_info = PyObject_GetAttrString(module, "get_cls_info");
     if (!get_cls_info)
         return 0;
@@ -241,7 +252,8 @@ initialize_globals(void)
     parenthesis_format = PyUnicode_DecodeASCII("(%s)", 4, "strict");
     default_compile_join = PyUnicode_DecodeASCII(", ", 2, "strict");
 
-    return 1;
+    initialized = 1;
+    return initialized;
 }
 
 
@@ -362,7 +374,7 @@ EventSystem_unhook(EventSystemObject *self, PyObject *args)
     callback = PyTuple_GET_ITEM(args, 1);
     data = PyTuple_GetSlice(args, 2, PyTuple_GET_SIZE(args));
     if (data) {
-        /* 
+        /*
            callbacks = self._hooks.get(name)
            if callbacks is not None:
                callbacks.discard((callback, data))
@@ -458,7 +470,7 @@ EventSystem_emit(EventSystemObject *self, PyObject *all_args)
                         PyObject *callback = PyTuple_GET_ITEM(item, 0);
                         PyObject *data = PyTuple_GET_ITEM(item, 1);
                         PyObject *res;
-                        /* 
+                        /*
                            if callback(owner, *(args+data)) is False:
                                callbacks.discard((callback, data))
                         */
@@ -527,8 +539,8 @@ statichere PyTypeObject EventSystem_Type = {
     0,                      /*tp_hash*/
     0,                      /*tp_call*/
     0,                      /*tp_str*/
-    PyObject_GenericGetAttr,/*tp_getattro*/
-    PyObject_GenericSetAttr,/*tp_setattro*/
+    0,                      /*tp_getattro*/
+    0,                      /*tp_setattro*/
     0,                      /*tp_as_buffer*/
     Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE | Py_TPFLAGS_HAVE_GC, /*tp_flags*/
     0,                      /*tp_doc*/
@@ -547,9 +559,9 @@ statichere PyTypeObject EventSystem_Type = {
     0,                      /*tp_descr_set*/
     0,                      /*tp_dictoffset*/
     (initproc)EventSystem_init, /*tp_init*/
-    PyType_GenericAlloc,    /*tp_alloc*/
-    PyType_GenericNew,      /*tp_new*/
-    PyObject_GC_Del,        /*tp_free*/
+    0,                      /*tp_alloc*/
+    0,                      /*tp_new*/
+    0,                      /*tp_free*/
     0,                      /*tp_is_gc*/
 };
 
@@ -698,30 +710,6 @@ Variable_dealloc(VariableObject *self)
     self->ob_type->tp_free((PyObject *)self);
 }
 
-static long
-Variable_hash(VariableObject *self)
-{
-    /* return hash(self._value) */
-    return PyObject_Hash((PyObject *)self->_value);
-}
-
-static PyObject *
-Variable_richcompare(VariableObject *self, VariableObject *other, int op)
-{
-    /*
-       return (self.__class__ is other.__class__ and
-               self._value == other._value)
-    */
-    /* This test will also prevent that we access _value on a
-       non-Variable object. */
-    if (op == Py_EQ &&
-        ((PyObject *)self)->ob_type != ((PyObject *)other)->ob_type) {
-        Py_INCREF(Py_False);
-        return Py_False;
-    }
-    return PyObject_RichCompare(self->_value, other->_value, op);
-}
-
 static PyObject *
 Variable_parse_get(VariableObject *self, PyObject *args)
 {
@@ -755,7 +743,7 @@ Variable_get_lazy(VariableObject *self, PyObject *args, PyObject *kwargs)
                                      &default_))
         return NULL;
 
-    /* 
+    /*
        if self._lazy_value is Undef:
            return default
        return self._lazy_value
@@ -834,9 +822,12 @@ Variable_set(VariableObject *self, PyObject *args, PyObject *kwargs)
         Py_INCREF(value);
         REPLACE(self->_lazy_value, value);
 
-        /* new_value = Undef */
+        /* self._checkpoint_state = new_value = Undef */
+        Py_INCREF(Undef);
         Py_INCREF(Undef);
         new_value = Undef;
+        Py_DECREF(self->_checkpoint_state);
+        self->_checkpoint_state = Undef;
     }
     /* else: */
     else {
@@ -1140,17 +1131,17 @@ statichere PyTypeObject Variable_Type = {
     0,            /*tp_as_number*/
     0,            /*tp_as_sequence*/
     0,            /*tp_as_mapping*/
-    (hashfunc)Variable_hash, /*tp_hash*/
+    0,            /*tp_hash*/
     0,                      /*tp_call*/
     0,                      /*tp_str*/
-    PyObject_GenericGetAttr,/*tp_getattro*/
-    PyObject_GenericSetAttr,/*tp_setattro*/
+    0,                      /*tp_getattro*/
+    0,                      /*tp_setattro*/
     0,                      /*tp_as_buffer*/
     Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
     0,                      /*tp_doc*/
     (traverseproc)Variable_traverse,  /*tp_traverse*/
     (inquiry)Variable_clear,          /*tp_clear*/
-    (richcmpfunc)Variable_richcompare, /*tp_richcompare*/
+    0,                      /*tp_richcompare*/
     0,                      /*tp_weaklistoffset*/
     0,                      /*tp_iter*/
     0,                      /*tp_iternext*/
@@ -1163,9 +1154,9 @@ statichere PyTypeObject Variable_Type = {
     0,                      /*tp_descr_set*/
     0,                      /*tp_dictoffset*/
     (initproc)Variable_init, /*tp_init*/
-    PyType_GenericAlloc,    /*tp_alloc*/
+    0,                       /*tp_alloc*/
     Variable_new,      /*tp_new*/
-    PyObject_GC_Del,        /*tp_free*/
+    0,                      /*tp_free*/
     0,                      /*tp_is_gc*/
 };
 
@@ -1346,7 +1337,7 @@ Compile_when(CompileObject *self, PyObject *types)
         }
         Py_DECREF(module);
     }
-    return result;    
+    return result;
 }
 
 static PyObject *
@@ -1430,7 +1421,7 @@ Compile_is_reserved_word(CompileObject *self, PyObject *word)
     CATCH(NULL, lower_word = PyObject_CallMethod(word, "lower", NULL));
     item = PyDict_GetItem(self->_reserved_words, lower_word);
     if (item == NULL && PyErr_Occurred()) {
-        goto error; 
+        goto error;
     } else if (item != NULL && item != Py_None) {
         result = Py_True;
     }
@@ -1609,7 +1600,7 @@ Compile_one_or_many(CompileObject *self, PyObject *expr, PyObject *state,
       if (expr_type is SQLRaw or
           raw and (expr_type is str or expr_type is unicode)):
           return expr
-    */  
+    */
     if ((PyObject *)expr->ob_type == SQLRaw ||
         (raw && (PyString_CheckExact(expr) || PyUnicode_CheckExact(expr)))) {
         /* Pass our reference on. */
@@ -1683,7 +1674,7 @@ Compile_one_or_many(CompileObject *self, PyObject *expr, PyObject *state,
                     Py_INCREF(subexpr);
                 }
 
-                /* 
+                /*
                    statement = self._compile_single(subexpr, state,
                                                     outer_precedence)
                 */
@@ -1692,7 +1683,7 @@ Compile_one_or_many(CompileObject *self, PyObject *expr, PyObject *state,
                 Py_DECREF(subexpr);
                 CATCH(NULL, statement);
             }
-            
+
             /* compiled.append(statement) */
             CATCH(-1, PyList_Append(compiled, statement));
             Py_CLEAR(statement);
@@ -1807,8 +1798,8 @@ statichere PyTypeObject Compile_Type = {
     0,                      /*tp_hash*/
     (ternaryfunc)Compile__call__, /*tp_call*/
     0,                      /*tp_str*/
-    PyObject_GenericGetAttr,/*tp_getattro*/
-    PyObject_GenericSetAttr,/*tp_setattro*/
+    0,                      /*tp_getattro*/
+    0,                      /*tp_setattro*/
     0,                      /*tp_as_buffer*/
     Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE | Py_TPFLAGS_HAVE_GC, /*tp_flags*/
     0,                      /*tp_doc*/
@@ -1827,9 +1818,9 @@ statichere PyTypeObject Compile_Type = {
     0,                      /*tp_descr_set*/
     0,                      /*tp_dictoffset*/
     (initproc)Compile_init, /*tp_init*/
-    PyType_GenericAlloc,    /*tp_alloc*/
-    PyType_GenericNew,      /*tp_new*/
-    PyObject_GC_Del,        /*tp_free*/
+    0,                      /*tp_alloc*/
+    0,                      /*tp_new*/
+    0,                      /*tp_free*/
     0,                      /*tp_is_gc*/
 };
 
@@ -2021,6 +2012,26 @@ ObjectInfo_clear(ObjectInfoObject *self)
     return PyDict_Type.tp_clear((PyObject *)self);
 }
 
+static PyObject *
+ObjectInfo_richcompare(PyObject *self, PyObject *other, int op)
+{
+    PyObject *res;
+
+    /* Implement equality via object identity. */
+    switch (op) {
+    case Py_EQ:
+        res = (self == other) ? Py_True : Py_False;
+        break;
+    case Py_NE:
+        res = (self != other) ? Py_True : Py_False;
+        break;
+    default:
+        res = Py_NotImplemented;
+    }
+    Py_INCREF(res);
+    return res;
+}
+
 static void
 ObjectInfo_dealloc(ObjectInfoObject *self)
 {
@@ -2075,17 +2086,17 @@ statichere PyTypeObject ObjectInfo_Type = {
     0,            /*tp_as_number*/
     0,            /*tp_as_sequence*/
     0,            /*tp_as_mapping*/
-    (hashfunc)_Py_HashPointer, /*tp_hash*/
+    0,                      /*tp_hash*/
     0,                      /*tp_call*/
     0,                      /*tp_str*/
-    PyObject_GenericGetAttr, /*tp_getattro*/
-    PyObject_GenericSetAttr, /*tp_setattro*/
+    0,                      /*tp_getattro*/
+    0,                      /*tp_setattro*/
     0,                      /*tp_as_buffer*/
     Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
     0,                      /*tp_doc*/
     (traverseproc)ObjectInfo_traverse, /*tp_traverse*/
     (inquiry)ObjectInfo_clear, /*tp_clear*/
-    0,                      /*tp_richcompare*/
+    ObjectInfo_richcompare, /*tp_richcompare*/
     offsetof(ObjectInfoObject, __weakreflist), /*tp_weaklistoffset*/
     0,                      /*tp_iter*/
     0,                      /*tp_iternext*/
@@ -2098,9 +2109,9 @@ statichere PyTypeObject ObjectInfo_Type = {
     0,                      /*tp_descr_set*/
     0,                      /*tp_dictoffset*/
     (initproc)ObjectInfo_init, /*tp_init*/
-    PyType_GenericAlloc,    /*tp_alloc*/
+    0,                      /*tp_alloc*/
     0,                      /*tp_new*/
-    PyObject_GC_Del,        /*tp_free*/
+    0,                      /*tp_free*/
     0,                      /*tp_is_gc*/
 };
 
@@ -2146,17 +2157,36 @@ static PyMethodDef cextensions_methods[] = {
 };
 
 
+static int
+prepare_type(PyTypeObject *type)
+{
+    if (!type->tp_getattro && !type->tp_getattr)
+        type->tp_getattro = PyObject_GenericGetAttr;
+    if (!type->tp_setattro && !type->tp_setattr)
+        type->tp_setattro = PyObject_GenericSetAttr;
+    if (!type->tp_alloc)
+        type->tp_alloc = PyType_GenericAlloc;
+    /* Don't fill in tp_new if this class has a base class */
+    if (!type->tp_base && !type->tp_new)
+        type->tp_new = PyType_GenericNew;
+    if (!type->tp_free) {
+        assert((type->tp_flags & Py_TPFLAGS_HAVE_GC) != 0);
+        type->tp_free = PyObject_GC_Del;
+    }
+    return PyType_Ready(type);
+}
+
 DL_EXPORT(void)
 initcextensions(void)
 {
     PyObject *module;
 
-    PyType_Ready(&EventSystem_Type);
-    PyType_Ready(&Variable_Type);
-    PyType_Ready(&Compile_Type);
-
+    prepare_type(&EventSystem_Type);
+    prepare_type(&Compile_Type);
     ObjectInfo_Type.tp_base = &PyDict_Type;
-    PyType_Ready(&ObjectInfo_Type);
+    ObjectInfo_Type.tp_hash = (hashfunc)_Py_HashPointer;
+    prepare_type(&ObjectInfo_Type);
+    prepare_type(&Variable_Type);
 
     module = Py_InitModule3("cextensions", cextensions_methods, "");
     Py_INCREF(&Variable_Type);
