@@ -18,11 +18,14 @@
 # You should have received a copy of the GNU Lesser General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #
-from storm.exceptions import WrongStoreError, NoStoreError, ClassInfoError
-from storm.store import Store, get_where_for_args
+import weakref
+
+from storm.exceptions import (
+    ClassInfoError, FeatureError, NoStoreError, WrongStoreError)
+from storm.store import Store, get_where_for_args, LostObjectError
 from storm.variables import LazyValue
 from storm.expr import (
-    Select, Column, Exists, ComparableExpr, LeftJoin, SQLRaw,
+    Select, Column, Exists, ComparableExpr, SuffixExpr, LeftJoin, Not, SQLRaw,
     compare_columns, compile)
 from storm.info import get_cls_info, get_obj_info
 
@@ -200,6 +203,9 @@ class Reference(object):
     def __eq__(self, other):
         return self._relation.get_where_for_local(other)
 
+    def __ne__(self, other):
+        return Not(self == other)
+
 
 class ReferenceSet(object):
 
@@ -208,6 +214,7 @@ class ReferenceSet(object):
     # not be available yet.
     _relation1 = LazyAttribute("_relation1", "_build_relations")
     _relation2 = LazyAttribute("_relation2", "_build_relations")
+    _order_by = LazyAttribute("_order_by", "_build_relations")
 
     def __init__(self, local_key1, remote_key1,
                  remote_key2=None, local_key2=None, order_by=None):
@@ -215,7 +222,7 @@ class ReferenceSet(object):
         self._remote_key1 = remote_key1
         self._remote_key2 = remote_key2
         self._local_key2 = local_key2
-        self._order_by = order_by
+        self._default_order_by = order_by
         self._cls = None
 
     def __get__(self, local, cls=None):
@@ -240,9 +247,17 @@ class ReferenceSet(object):
                                              self._relation2, local,
                                              self._order_by)
 
+    def __set__(self, local, value):
+        raise FeatureError("Assigning to ResultSets not supported")
+
     def _build_relations(self):
         resolver = PropertyResolver(self, self._cls)
 
+        if self._default_order_by is not None:
+            self._order_by = resolver.resolve(self._default_order_by)
+        else:
+            self._order_by = None
+
         self._local_key1 = resolver.resolve(self._local_key1)
         self._remote_key1 = resolver.resolve(self._remote_key1)
         self._relation1 = Relation(self._local_key1, self._remote_key1,
@@ -266,7 +281,7 @@ class BoundReferenceSetBase(object):
         where = self._get_where_clause()
         result = store.find(self._target_cls, where, *args, **kwargs)
         if self._order_by is not None:
-            result.order_by(self._order_by)
+            result.order_by(*self._order_by)
         return result
 
     def __iter__(self):
@@ -462,11 +477,23 @@ class Relation(object):
         self._r_to_l = {}
 
     def get_remote(self, local):
+        """Return the remote object for this relation, using the local cache.
+
+        If the object in the cache is invalidated, we validate it again to
+        check if it's still in the database.
+        """
         local_info = get_obj_info(local)
         try:
-            return local_info[self]["remote"]
+            obj = local_info[self]["remote"]
         except KeyError:
             return None
+        remote_info = get_obj_info(obj)
+        if remote_info.get("invalidated"):
+            try:
+                Store.of(obj)._validate_alive(remote_info)
+            except LostObjectError:
+                return None
+        return obj
 
     def get_where_for_remote(self, local):
         """Generate a column comparison expression for reference properties.
@@ -617,7 +644,7 @@ class Relation(object):
                 #local_info.event.hook("removed", self._break_on_local_removed,
                 #                      remote_info)
                 remote_info.event.hook("removed", self._break_on_remote_removed,
-                                       local_info)
+                                       weakref.ref(local_info))
             else:
                 remote_has_changed = False
                 for local_column, remote_column in pairs:
@@ -646,10 +673,10 @@ class Relation(object):
             local_info.event.hook("changed", self._break_on_local_diverged,
                                   remote_info)
             remote_info.event.hook("changed", self._break_on_remote_diverged,
-                                   local_info)
+                                   weakref.ref(local_info))
             if self.on_remote:
                 remote_info.event.hook("removed", self._break_on_remote_removed,
-                                       local_info)
+                                       weakref.ref(local_info))
 
     def unlink(self, local_info, remote_info, setting=False):
         """Break the relation between the local and remote objects.
@@ -681,11 +708,11 @@ class Relation(object):
             remote_info.event.unhook("changed", self._track_remote_changes,
                                      local_info)
             remote_info.event.unhook("changed", self._break_on_remote_diverged,
-                                     local_info)
+                                     weakref.ref(local_info))
             remote_info.event.unhook("flushed", self._break_on_remote_flushed,
                                      local_info)
             remote_info.event.unhook("removed", self._break_on_remote_removed,
-                                     local_info)
+                                     weakref.ref(local_info))
 
             if local_store is None:
                 if not self.many or not remote_infos:
@@ -778,13 +805,16 @@ class Relation(object):
                 self.unlink(local_info, remote_info)
 
     def _break_on_remote_diverged(self, remote_info, remote_variable,
-                                  old_value, new_value, fromdb, local_info):
+                                  old_value, new_value, fromdb, local_info_ref):
         """Break the remote/local relationship on diverging changes.
 
         This hook ensures that if the remote object has an attribute
         changed by hand in a way that diverges from the local object,
         the relationship is undone.
         """
+        local_info = local_info_ref()
+        if local_info is None:
+            return
         local_column = self._get_local_column(local_info.cls_info.cls,
                                               remote_variable.column)
         if local_column is not None:
@@ -800,9 +830,11 @@ class Relation(object):
         """Break the remote/local relationship on flush."""
         self.unlink(local_info, remote_info)
 
-    def _break_on_remote_removed(self, remote_info, local_info):
+    def _break_on_remote_removed(self, remote_info, local_info_ref):
         """Break the remote relationship when the remote object is removed."""
-        self.unlink(local_info, remote_info)
+        local_info = local_info_ref()
+        if local_info is not None:
+            self.unlink(local_info, remote_info)
 
     def _add_all(self, obj_info, local_info):
         store = Store.of(obj_info)
@@ -880,6 +912,12 @@ class PropertyResolver(object):
             return self.resolve(property)
         elif isinstance(property, basestring):
             return self._resolve_string(property)
+        elif isinstance(property, SuffixExpr):
+            # XXX This covers cases like order_by=Desc("Bar.id"), see #620369.
+            # Eventually we might want to add support for other types of
+            # expressions
+            property.expr = self.resolve(property.expr)
+            return property
         elif not isinstance(property, Column):
             return _find_descriptor_obj(self._used_cls, property)
         return property
@@ -887,7 +925,7 @@ class PropertyResolver(object):
     def _resolve_string(self, property_path):
         if self._registry is None:
             try:
-                registry = self._used_cls._storm_property_registry
+                self._registry = self._used_cls._storm_property_registry
             except AttributeError:
                 raise RuntimeError("When using strings on references, "
                                    "classes involved must be subclasses "
@@ -895,7 +933,7 @@ class PropertyResolver(object):
             cls = _find_descriptor_class(self._used_cls, self._reference)
             self._namespace = "%s.%s" % (cls.__module__, cls.__name__)
 
-        return registry.get(property_path, self._namespace)
+        return self._registry.get(property_path, self._namespace)
 
 
 def _find_descriptor_class(used_cls, descr):
