@@ -5,7 +5,7 @@
 """
 
 #
-# Copyright (c) 2006, 2007 Canonical
+# Copyright (c) 2006-2009 Canonical
 #
 # Written by Gustavo Niemeyer <gustavo@niemeyer.net>
 #
@@ -27,10 +27,12 @@
 import threading
 import weakref
 
+from uuid import uuid4
+
 from zope.interface import implements
 
 import transaction
-from transaction.interfaces import IDataManager, ISynchronizer
+from transaction.interfaces import IDataManager
 try:
     from transaction.interfaces import TransactionFailedError
 except ImportError:
@@ -39,6 +41,7 @@ except ImportError:
 from storm.zope.interfaces import IZStorm, ZStormError
 from storm.database import create_database
 from storm.store import Store
+from storm.xid import Xid
 
 
 class ZStorm(object):
@@ -60,12 +63,15 @@ class ZStorm(object):
 
     implements(IZStorm)
 
+    transaction_manager = transaction.manager
+
     _databases = {}
 
     def __init__(self):
         self._local = threading.local()
         self._default_databases = {}
         self._default_uris = {}
+        self._default_tpcs = {}
 
     def _reset(self):
         for name, store in list(self.iterstores()):
@@ -75,6 +81,7 @@ class ZStorm(object):
         self._databases.clear()
         self._default_databases.clear()
         self._default_uris.clear()
+        self._default_tpcs.clear()
 
     @property
     def _stores(self):
@@ -99,6 +106,17 @@ class ZStorm(object):
             return self._local.__dict__.setdefault(
                 "name_index", weakref.WeakKeyDictionary())
 
+    @property
+    def _txn_ids(self):
+        """
+        A thread-local weak-key dict used to keep track of transaction IDs.
+        """
+        try:
+            return self._local.txn_ids
+        except AttributeError:
+            txn_ids = weakref.WeakKeyDictionary()
+            return self._local.__dict__.setdefault("txn_ids", txn_ids)
+
     def _get_database(self, uri):
         database = self._databases.get(uri)
         if database is None:
@@ -110,6 +128,10 @@ class ZStorm(object):
         self._default_databases[name] = self._get_database(default_uri)
         self._default_uris[name] = default_uri
 
+    def set_default_tpc(self, name, default_flag):
+        """Set the default two-phase mode for stores with the given C{name}."""
+        self._default_tpcs[name] = default_flag
+
     def create(self, name, uri=None):
         """Create a new store called C{name}.
 
@@ -129,7 +151,11 @@ class ZStorm(object):
             raise ZStormError("Store named '%s' already exists" % name)
 
         store = Store(database)
-        store.__synchronizer = StoreSynchronizer(store)
+        store._register_for_txn = True
+        store._tpc = self._default_tpcs.get(name, False)
+        store._event.hook(
+            "register-transaction", register_store_with_transaction,
+            weakref.ref(self))
 
         self._stores[id(store)] = store
         if name is not None:
@@ -139,10 +165,12 @@ class ZStorm(object):
         return store
 
     def get(self, name, default_uri=None):
-        """Get the store called C{name} or None if one isn't available.
+        """Get the store called C{name}, creating it first if necessary.
 
         @param default_uri: Optionally, the URI to use to create a
            store called C{name} when one doesn't already exist.
+        @raises ZStormError: Raised if C{uri} is None and no default
+            URI exists for C{name}.
         """
         store = self._named.get(name)
         if not store:
@@ -168,7 +196,12 @@ class ZStorm(object):
         del self._name_index[store]
         if name in self._named:
             del self._named[name]
-        transaction.manager.unregisterSynch(store.__synchronizer)
+
+        # Make sure the store isn't hooked up to future transactions.
+        store._register_for_txn = False
+        store._event.unhook(
+            "register-transaction", register_store_with_transaction,
+            weakref.ref(self))
 
     def iterstores(self):
         """Iterate C{name, store} 2-tuples."""
@@ -191,53 +224,35 @@ class ZStorm(object):
         return self._default_uris.copy()
 
 
-class StoreSynchronizer(object):
-    """This class takes cares of plugging the store in new transactions.
+def register_store_with_transaction(store, zstorm_ref):
+    zstorm = zstorm_ref()
+    if zstorm is None:
+        # zstorm object does not exist any more.
+        return False
 
-    Garbage collection should work fine, because the transaction manager
-    stores synchronizers as weak references. Even then, we give a hand
-    to the garbage collector by avoiding strong circular references
-    on the store.
-    """
+    # Check if the store is  known.  This could indicate a store being
+    # used outside of its thread.
+    if id(store) not in zstorm._stores:
+        raise ZStormError("Store not registered with ZStorm, or registered "
+                          "with another thread.")
 
-    implements(ISynchronizer)
+    txn = zstorm.transaction_manager.get()
 
-    def __init__(self, store):
-        data_manager = StoreDataManager(store)
+    if store._tpc:
+        global_transaction_id = zstorm._txn_ids.get(txn)
+        if global_transaction_id is None:
+            # The the global transaction doesn't have an ID yet, let's create
+            # one in a way that it will be unique
+            global_transaction_id = "_storm_%s" % str(uuid4())
+            zstorm._txn_ids[txn] = global_transaction_id
+        xid = Xid(0, global_transaction_id, zstorm.get_name(store))
+        store.begin(xid)
 
-        self._store_ref = weakref.ref(store)
-        self._data_manager_ref = weakref.ref(data_manager)
+    data_manager = StoreDataManager(store, zstorm)
+    txn.join(data_manager)
 
-        # Join now ...
-        data_manager.join(transaction.get())
-
-        # ... and in the future.
-        transaction.manager.registerSynch(self)
-
-    def _join(self, trans):
-        # If the store is still alive and the transaction is in this thread.
-        store = self._store_ref()
-        if store and trans is transaction.get():
-            data_manager = self._data_manager_ref()
-            if data_manager is None:
-                data_manager = StoreDataManager(store)
-                self._data_manager_ref = weakref.ref(data_manager)
-            try:
-                data_manager.join(trans)
-            except TransactionFailedError:
-                # It means that an *already failed* transaction is trying
-                # to join us to notify that it is indeed failed.  We don't
-                # care about these (see the double_abort test case).
-                pass
-
-    def beforeCompletion(self, trans):
-        self._join(trans)
-
-    def afterCompletion(self, trans):
-        pass
-
-    def newTransaction(self, trans):
-        self._join(trans)
+    # Unhook the event handler.  It will be rehooked for the next transaction.
+    return False
 
 
 class StoreDataManager(object):
@@ -245,20 +260,25 @@ class StoreDataManager(object):
 
     implements(IDataManager)
 
-    transaction_manager = transaction.manager
-
-    def __init__(self, store):
+    def __init__(self, store, zstorm):
         self._store = store
-        self._trans = None
+        self._zstorm = zstorm
+        self.transaction_manager = zstorm.transaction_manager
 
-    def join(self, trans):
-        if trans is not self._trans:
-            self._trans = trans
-            trans.join(self)
+    def _hook_register_transaction_event(self):
+        if self._store._register_for_txn:
+            self._store._event.hook(
+                "register-transaction", register_store_with_transaction,
+                weakref.ref(self._zstorm))
 
     def abort(self, txn):
-        self._trans = None
-        self._store.rollback()
+        try:
+            self._store.rollback()
+        finally:
+            if self._store._register_for_txn:
+                self._store._event.hook(
+                    "register-transaction", register_store_with_transaction,
+                    weakref.ref(self._zstorm))
 
     def tpc_begin(self, txn):
         # Zope's transaction system will call tpc_begin() on all
@@ -267,23 +287,45 @@ class StoreDataManager(object):
         # of them will fail.  In such cases, flushing earlier will
         # ensure that both transactions will be rolled back, instead
         # of one committed and one rolled back.
+        #
+        # If TPC support is on, we still want to perform this flush for a
+        # couple of reasons. Firstly the queries flush() runs couldn't
+        # be run after calling prepare(), because the transaction is frozen
+        # waiting for the final commit(), and secondly because if the flush
+        # fails the entire transaction will be aborted with a normal rollback
+        # as opposed to a TPC rollback, that would happen after prepare().
         self._store.flush()
 
     def commit(self, txn):
-        self._trans = None
-        self._store.commit()
+        if self._store._tpc:
+            self._store.prepare()
 
     def tpc_vote(self, txn):
         pass
 
     def tpc_finish(self, txn):
-        pass
+        # If commit raises an exception, we let the exception propagate, as
+        # the transaction manager will then call tcp_abort, and we will
+        # register the hook there
+        self._store.commit()
+        self._hook_register_transaction_event()
 
     def tpc_abort(self, txn):
-        pass
+        if self._store._tpc:
+            try:
+                self._store.rollback()
+            finally:
+                self._hook_register_transaction_event()
 
     def sortKey(self):
-        return "store_%d" % id(self)
+        # Stores in TPC mode should be the last to be committed, this makes
+        # it possible to have TPC behavior when there's only a single store
+        # not in TPC mode.
+        if self._store._tpc:
+            prefix = "zz"
+        else:
+            prefix = "aa"
+        return "%s_store_%d" % (prefix, id(self))
 
 
 global_zstorm = ZStorm()
