@@ -118,6 +118,12 @@ class FakeTracer(object):
         self.seen.append(("ERROR", connection, type(raw_cursor),
                           statement, params, error))
 
+    def connection_commit(self, connection, xid=None):
+        self.seen.append(("COMMIT", connection, xid))
+
+    def connection_rollback(self, connection, xid=None):
+        self.seen.append(("ROLLBACK", connection, xid))
+
 
 class DatabaseTest(TestHelper):
 
@@ -158,10 +164,6 @@ class ConnectionTest(TestHelper):
         self.assertEquals(result, None)
         self.assertEquals(self.executed, [("something", marker), "RCLOSE"])
 
-    def test_execute_convert_param_style(self):
-        self.connection.execute("'?' ? '?' ? '?'")
-        self.assertEquals(self.executed, [("'?' ? '?' ? '?'", marker)])
-
     def test_execute_convert_param_style(self):
         class MyConnection(Connection):
             param_mark = "%s"
@@ -227,6 +229,24 @@ class ConnectionTest(TestHelper):
                                         ("ERROR", self.connection, RawCursor,
                                          "something", (), error)])
 
+    def test_raw_execute_setup_error_tracing(self):
+        """
+        When an exception is raised in the connection_raw_execute hook of a
+        tracer, the connection_raw_execute_error hook is called.
+        """
+        cursor_mock = self.mocker.patch(FakeTracer)
+        cursor_mock.connection_raw_execute(ARGS)
+        error = ZeroDivisionError()
+        self.mocker.throw(error)
+        self.mocker.replay()
+
+        tracer = FakeTracer()
+        install_tracer(tracer)
+        self.assertRaises(ZeroDivisionError,
+                          self.connection.execute, "something")
+        self.assertEquals(tracer.seen, [("ERROR", self.connection, RawCursor,
+                                         "something", (), error)])
+
     def test_tracing_check_disconnect(self):
         tracer = FakeTracer()
         tracer_mock = self.mocker.patch(tracer)
@@ -282,10 +302,24 @@ class ConnectionTest(TestHelper):
         self.connection.commit()
         self.assertEquals(self.executed, ["COMMIT"])
 
+    def test_commit_tracing(self):
+        self.assertMethodsMatch(FakeTracer, DebugTracer)
+        tracer = FakeTracer()
+        install_tracer(tracer)
+        self.connection.commit()
+        self.assertEquals(tracer.seen, [("COMMIT", self.connection, None)])
+
     def test_rollback(self):
         self.connection.rollback()
         self.assertEquals(self.executed, ["ROLLBACK"])
 
+    def test_rollback_tracing(self):
+        self.assertMethodsMatch(FakeTracer, DebugTracer)
+        tracer = FakeTracer()
+        install_tracer(tracer)
+        self.connection.rollback()
+        self.assertEquals(tracer.seen, [("ROLLBACK", self.connection, None)])
+
     def test_close(self):
         self.connection.close()
         self.assertEquals(self.executed, ["CCLOSE"])
