@@ -22,26 +22,54 @@ from datetime import date, time, timedelta
 import os
 
 from storm.databases.postgres import (
-    Postgres, compile, currval, Returning, PostgresTimeoutTracer)
+    Postgres, compile, currval, Returning, PostgresTimeoutTracer, make_dsn)
 from storm.database import create_database
 from storm.exceptions import InterfaceError, ProgrammingError
 from storm.variables import DateTimeVariable, RawStrVariable
 from storm.variables import ListVariable, IntVariable, Variable
 from storm.properties import Int
-from storm.expr import (Union, Select, Insert, Alias, SQLRaw, State,
+from storm.exceptions import DisconnectionError
+from storm.expr import (Union, Select, Insert, Update, Alias, SQLRaw, State,
                         Sequence, Like, Column, COLUMN)
 from storm.tracer import install_tracer, TimeoutError
+from storm.uri import URI
 
 # We need the info to register the 'type' compiler.  In normal
 # circumstances this is naturally imported.
 import storm.info
+storm  # Silence lint.
 
+from tests import has_fixtures, has_subunit
 from tests.databases.base import (
-    DatabaseTest, DatabaseDisconnectionTest, UnsupportedDatabaseTest)
+    DatabaseTest, DatabaseDisconnectionTest, UnsupportedDatabaseTest,
+    TwoPhaseCommitTest, TwoPhaseCommitDisconnectionTest)
 from tests.expr import column1, column2, column3, elem1, table1, TrackContext
 from tests.tracer import TimeoutTracerTestBase
 from tests.helper import TestHelper
 
+try:
+    import pgbouncer
+except ImportError:
+    has_pgbouncer = False
+else:
+    has_pgbouncer = True
+
+
+def terminate_other_backends(connection):
+    """Terminate all connections to the database except the one given."""
+    connection.execute(
+        "SELECT pg_terminate_backend(procpid)"
+        "  FROM pg_stat_activity"
+        " WHERE datname = current_database()"
+        "   AND procpid != pg_backend_pid()")
+
+
+def terminate_all_backends(database):
+    """Terminate all connections to the given database."""
+    connection = database.connect()
+    terminate_other_backends(connection)
+    connection.close()
+
 
 class PostgresTest(DatabaseTest, TestHelper):
 
@@ -63,13 +91,13 @@ class PostgresTest(DatabaseTest, TestHelper):
                                 "(id SERIAL PRIMARY KEY, b BYTEA)")
         self.connection.execute("CREATE TABLE like_case_insensitive_test "
                                 "(id SERIAL PRIMARY KEY, description TEXT)")
-        self.connection.execute("CREATE TABLE insert_returning_test "
+        self.connection.execute("CREATE TABLE returning_test "
                                 "(id1 INTEGER DEFAULT 123, "
                                 " id2 INTEGER DEFAULT 456)")
 
     def drop_tables(self):
         super(PostgresTest, self).drop_tables()
-        for table in ["like_case_insensitive_test", "insert_returning_test"]:
+        for table in ["like_case_insensitive_test", "returning_test"]:
             try:
                 self.connection.execute("DROP TABLE %s" % table)
                 self.connection.commit()
@@ -362,25 +390,32 @@ class PostgresTest(DatabaseTest, TestHelper):
                           "thetable.thecolumn = "
                           "(SELECT currval('thetable_thecolumn_seq'))")
 
-    def test_returning(self):
-        insert = Insert({column1: elem1}, table1,
-                        primary_columns=(column2, column3))
-        self.assertEquals(compile(Returning(insert)),
-                          'INSERT INTO "table 1" (column1) VALUES (elem1) '
-                          'RETURNING column2, column3')
-
     def test_returning_column_context(self):
         column2 = TrackContext()
         insert = Insert({column1: elem1}, table1, primary_columns=column2)
         compile(Returning(insert))
         self.assertEquals(column2.context, COLUMN)
 
+    def test_returning_update(self):
+        update = Update({column1: elem1}, table=table1,
+                        primary_columns=(column2, column3))
+        self.assertEquals(compile(Returning(update)),
+                          'UPDATE "table 1" SET column1=elem1 '
+                          'RETURNING column2, column3')
+
+    def test_returning_update_with_columns(self):
+        update = Update({column1: elem1}, table=table1,
+                        primary_columns=(column2, column3))
+        self.assertEquals(compile(Returning(update, columns=[column3])),
+                          'UPDATE "table 1" SET column1=elem1 '
+                          'RETURNING column3')
+
     def test_execute_insert_returning(self):
-        if self.database._version < (8, 2):
+        if self.database._version < 80200:
             return # Can't run this test with old PostgreSQL versions.
 
-        column1 = Column("id1", "insert_returning_test")
-        column2 = Column("id2", "insert_returning_test")
+        column1 = Column("id1", "returning_test")
+        column2 = Column("id2", "returning_test")
         variable1 = IntVariable()
         variable2 = IntVariable()
         insert = Insert({}, primary_columns=(column1, column2),
@@ -393,13 +428,13 @@ class PostgresTest(DatabaseTest, TestHelper):
         self.assertEquals(variable1.get(), 123)
         self.assertEquals(variable2.get(), 456)
 
-        result = self.connection.execute("SELECT * FROM insert_returning_test")
+        result = self.connection.execute("SELECT * FROM returning_test")
         self.assertEquals(result.get_one(), (123, 456))
 
     def test_wb_execute_insert_returning_not_used_with_old_postgres(self):
         """Shouldn't try to use RETURNING with PostgreSQL < 8.2."""
-        column1 = Column("id1", "insert_returning_test")
-        column2 = Column("id2", "insert_returning_test")
+        column1 = Column("id1", "returning_test")
+        column2 = Column("id2", "returning_test")
         variable1 = IntVariable()
         variable2 = IntVariable()
         insert = Insert({}, primary_columns=(column1, column2),
@@ -411,31 +446,44 @@ class PostgresTest(DatabaseTest, TestHelper):
         self.assertFalse(variable1.is_defined())
         self.assertFalse(variable2.is_defined())
 
-        result = self.connection.execute("SELECT * FROM insert_returning_test")
+        result = self.connection.execute("SELECT * FROM returning_test")
         self.assertEquals(result.get_one(), (123, 456))
 
     def test_execute_insert_returning_without_columns(self):
         """Without primary_columns, the RETURNING system won't be used."""
-        column1 = Column("id1", "insert_returning_test")
+        column1 = Column("id1", "returning_test")
         variable1 = IntVariable()
         insert = Insert({column1: 123}, primary_variables=(variable1,))
         self.connection.execute(insert)
 
         self.assertFalse(variable1.is_defined())
 
-        result = self.connection.execute("SELECT * FROM insert_returning_test")
+        result = self.connection.execute("SELECT * FROM returning_test")
         self.assertEquals(result.get_one(), (123, 456))
 
     def test_execute_insert_returning_without_variables(self):
         """Without primary_variables, the RETURNING system won't be used."""
-        column1 = Column("id1", "insert_returning_test")
+        column1 = Column("id1", "returning_test")
         insert = Insert({}, primary_columns=(column1,))
         self.connection.execute(insert)
 
-        result = self.connection.execute("SELECT * FROM insert_returning_test")
+        result = self.connection.execute("SELECT * FROM returning_test")
 
         self.assertEquals(result.get_one(), (123, 456))
 
+    def test_execute_update_returning(self):
+        if self.database._version < 80200:
+            return # Can't run this test with old PostgreSQL versions.
+
+        column1 = Column("id1", "returning_test")
+        column2 = Column("id2", "returning_test")
+        self.connection.execute(
+            "INSERT INTO returning_test VALUES (1, 2)")
+        update = Update({"id2": 3}, column1 == 1,
+                        primary_columns=(column1, column2))
+        result = self.connection.execute(Returning(update))
+        self.assertEquals(result.get_one(), (1, 3))
+
     def test_isolation_autocommit(self):
         database = create_database(
             os.environ["STORM_POSTGRES_URI"] + "?isolation=autocommit")
@@ -506,13 +554,58 @@ class PostgresTest(DatabaseTest, TestHelper):
         connection.rollback()
 
     def test_default_isolation(self):
+        """
+        The default isolation level is REPEATABLE READ, but it's only supported
+        by psycopg2 2.4 and newer. Before, SERIALIZABLE is used instead.
+        """
         result = self.connection.execute("SHOW TRANSACTION ISOLATION LEVEL")
-        self.assertEquals(result.get_one()[0], u"serializable")
+        import psycopg2
+        psycopg2_version = psycopg2.__version__.split(None, 1)[0]
+        if psycopg2_version < "2.4":
+            self.assertEquals(result.get_one()[0], u"serializable")
+        else:
+            self.assertEquals(result.get_one()[0], u"repeatable read")
 
     def test_unknown_serialization(self):
         self.assertRaises(ValueError, create_database,
             os.environ["STORM_POSTGRES_URI"] + "?isolation=stuff")
 
+    def test_is_disconnection_error_with_ssl_syscall_error(self):
+        """
+        If the underlying driver raises a ProgrammingError with 'SSL SYSCALL
+        error', we consider the connection dead and mark it as needing
+        reconnection.
+        """
+        exc = ProgrammingError("SSL SYSCALL error: Connection timed out")
+        self.assertTrue(self.connection.is_disconnection_error(exc))
+
+
+_max_prepared_transactions = None
+
+
+class PostgresTwoPhaseCommitTest(TwoPhaseCommitTest, TestHelper):
+    
+    def is_supported(self):
+        uri = os.environ.get("STORM_POSTGRES_URI")
+        if not uri:
+            return False
+        global _max_prepared_transactions
+        if _max_prepared_transactions is None:
+            database = create_database(uri)
+            connection = database.connect()
+            result = connection.execute("SHOW MAX_PREPARED_TRANSACTIONS")
+            _max_prepared_transactions = int(result.get_one()[0])
+            connection.close()
+        return _max_prepared_transactions > 0
+
+    def create_database(self):
+        self.database = create_database(os.environ["STORM_POSTGRES_URI"])
+
+    def create_tables(self):
+        self.connection.execute("CREATE TABLE test "
+                                "(id SERIAL PRIMARY KEY, title VARCHAR)")
+        self.connection.commit()
+
 
 class PostgresUnsupportedTest(UnsupportedDatabaseTest, TestHelper):
 
@@ -520,7 +613,8 @@ class PostgresUnsupportedTest(UnsupportedDatabaseTest, TestHelper):
     db_module_name = "postgres"
 
 
-class PostgresDisconnectionTest(DatabaseDisconnectionTest, TestHelper):
+class PostgresDisconnectionTest(DatabaseDisconnectionTest, TwoPhaseCommitDisconnectionTest,
+                                TestHelper):
 
     environment_variable = "STORM_POSTGRES_URI"
     host_environment_variable = "STORM_POSTGRES_HOST_URI"
@@ -542,6 +636,112 @@ class PostgresDisconnectionTest(DatabaseDisconnectionTest, TestHelper):
             self.fail('Exception should have been swallowed: %s' % repr(exc))
 
 
+class PostgresDisconnectionTestWithoutProxyBase(object):
+    # DatabaseDisconnectionTest uses a socket proxy to simulate broken
+    # connections. This class tests some other causes of disconnection.
+
+    database_uri = None
+
+    def is_supported(self):
+        return bool(self.database_uri) and super(
+            PostgresDisconnectionTestWithoutProxyBase, self).is_supported()
+
+    def setUp(self):
+        super(PostgresDisconnectionTestWithoutProxyBase, self).setUp()
+        self.database = create_database(self.database_uri)
+
+    def test_terminated_backend(self):
+        # The error raised when trying to use a connection that has been
+        # terminated at the server is considered a disconnection error.
+        connection = self.database.connect()
+        terminate_all_backends(self.database)
+        self.assertRaises(
+            DisconnectionError, connection.execute,
+            "SELECT current_database()")
+
+
+if has_subunit:
+    # Some of the following tests are prone to segfaults, presumably in
+    # _psycopg.so. Run them in a subprocess if possible.
+    from subunit import IsolatedTestCase
+    class MisbehavingTestCase(TestHelper, IsolatedTestCase):
+        pass
+else:
+    # If we can't run them in a subprocess we still want to create tests, but
+    # prevent them from running, so that the skip is reported
+    class MisbehavingTestCase(TestHelper):
+        def is_supported(self):
+            return False
+
+
+class PostgresDisconnectionTestWithoutProxyUnixSockets(
+    PostgresDisconnectionTestWithoutProxyBase, MisbehavingTestCase):
+    """Disconnection tests using Unix sockets."""
+
+    database_uri = os.environ.get("STORM_POSTGRES_URI")
+
+class PostgresDisconnectionTestWithoutProxyTCPSockets(
+    PostgresDisconnectionTestWithoutProxyBase, MisbehavingTestCase):
+    """Disconnection tests using TCP sockets."""
+
+    database_uri = os.environ.get("STORM_POSTGRES_HOST_URI")
+
+
+class PostgresDisconnectionTestWithPGBouncerBase(object):
+    # Connecting via pgbouncer <http://pgfoundry.org/projects/pgbouncer>
+    # introduces new possible causes of disconnections.
+
+    def is_supported(self):
+        return (
+            has_fixtures and has_pgbouncer and
+            bool(os.environ.get("STORM_POSTGRES_HOST_URI")))
+
+    def setUp(self):
+        super(PostgresDisconnectionTestWithPGBouncerBase, self).setUp()
+        database_uri = URI(os.environ["STORM_POSTGRES_HOST_URI"])
+        database_user = database_uri.username or os.environ['USER']
+        database_dsn = make_dsn(database_uri)
+        # Create a pgbouncer fixture.
+        self.pgbouncer = pgbouncer.fixture.PGBouncerFixture()
+        self.pgbouncer.databases[database_uri.database] = database_dsn
+        self.pgbouncer.users[database_user] = "trusted"
+        self.pgbouncer.admin_users = [database_user]
+        self.useFixture(self.pgbouncer)
+        # Create a Database that uses pgbouncer.
+        pgbouncer_uri = database_uri.copy()
+        pgbouncer_uri.host = self.pgbouncer.host
+        pgbouncer_uri.port = self.pgbouncer.port
+        self.database = create_database(pgbouncer_uri)
+
+    def test_terminated_backend(self):
+        # The error raised when trying to use a connection through pgbouncer
+        # that has been terminated at the server is considered a disconnection
+        # error.
+        connection = self.database.connect()
+        terminate_all_backends(self.database)
+        self.assertRaises(
+            DisconnectionError, connection.execute,
+            "SELECT current_database()")
+
+    def test_pgbouncer_stopped(self):
+        # The error raised from a connection that is no longer connected
+        # because pgbouncer has been immediately shutdown (via SIGTERM; see
+        # man 1 pgbouncer) is considered a disconnection error.
+        connection = self.database.connect()
+        self.pgbouncer.stop()
+        self.assertRaises(
+            DisconnectionError, connection.execute,
+            "SELECT current_database()")
+
+
+if has_fixtures:
+    # Upgrade to full test case class with fixtures.
+    from fixtures import TestWithFixtures
+    class PostgresDisconnectionTestWithPGBouncer(
+        PostgresDisconnectionTestWithPGBouncerBase,
+        TestWithFixtures, TestHelper): pass
+
+
 class PostgresTimeoutTracerTest(TimeoutTracerTestBase):
 
     tracer_class = PostgresTimeoutTracer
