@@ -1,18 +1,30 @@
 import datetime
+import os
 import sys
 from unittest import TestCase
 
+from tests import has_fixtures
+
+# Optional dependency. If missing, Fixture tests are skipped.
+if has_fixtures:
+    import fixtures.testcase
+    TestWithFixtures = fixtures.testcase.TestWithFixtures
+    from storm.testing import CaptureTracer
+else:
+    TestWithFixtures = object
+
 try:
     # Optional dependency, if missing TimelineTracer tests are skipped.
     import timeline
+    has_timeline = True
 except ImportError:
-    timeline = None
+    has_timeline = False
 
-from storm.tracer import (trace, install_tracer, get_tracers,
+from storm.tracer import (trace, install_tracer, get_tracers, remove_tracer,
                           remove_tracer_type, remove_all_tracers, debug,
                           BaseStatementTracer, DebugTracer, TimeoutTracer,
                           TimelineTracer, TimeoutError, _tracers)
-from storm.database import Connection
+from storm.database import Connection, create_database
 from storm.expr import Variable
 
 from tests.helper import TestHelper
@@ -36,9 +48,28 @@ class TracerTest(TestHelper):
         remove_all_tracers()
         self.assertEquals(get_tracers(), [])
 
+    def test_remove_tracer(self):
+        """The C{remote_tracer} function removes a specific tracer."""
+        tracer1 = object()
+        tracer2 = object()
+        install_tracer(tracer1)
+        install_tracer(tracer2)
+        remove_tracer(tracer1)
+        self.assertEquals(get_tracers(), [tracer2])
+
+    def test_remove_tracer_with_not_installed_tracer(self):
+        """C{remote_tracer} exits gracefully if the tracer is not installed."""
+        tracer = object()
+        remove_tracer(tracer)
+        self.assertEquals(get_tracers(), [])
+
     def test_remove_tracer_type(self):
-        class C(object): pass
-        class D(C): pass
+        class C(object):
+            pass
+
+        class D(C):
+            pass
+
         c = C()
         d1 = D()
         d2 = D()
@@ -69,9 +100,11 @@ class TracerTest(TestHelper):
 
     def test_trace(self):
         stash = []
+
         class Tracer(object):
             def m1(_, *args, **kwargs):
                 stash.extend(["m1", args, kwargs])
+
             def m2(_, *args, **kwargs):
                 stash.extend(["m2", args, kwargs])
 
@@ -82,7 +115,6 @@ class TracerTest(TestHelper):
         self.assertEquals(stash, ["m1", (1, 2), {"c": 3}, "m2", (), {}])
 
 
-
 class MockVariable(Variable):
 
     def __init__(self, value):
@@ -101,7 +133,7 @@ class DebugTracerTest(TestHelper):
 
         datetime_mock = self.mocker.replace("datetime.datetime")
         datetime_mock.now()
-        self.mocker.result(datetime.datetime(1,2,3,4,5,6,7))
+        self.mocker.result(datetime.datetime(1, 2, 3, 4, 5, 6, 7))
         self.mocker.count(0, 1)
 
         self.variable = MockVariable("PARAM")
@@ -178,6 +210,24 @@ class DebugTracerTest(TestHelper):
         self.tracer.connection_raw_execute_success(connection, raw_cursor,
                                                    statement, params)
 
+    def test_connection_commit(self):
+        self.stream.write("[04:05:06.000007] COMMIT xid=None\n")
+        self.stream.flush()
+        self.mocker.replay()
+
+        connection = "CONNECTION"
+
+        self.tracer.connection_commit(connection)
+
+    def test_connection_rollback(self):
+        self.stream.write("[04:05:06.000007] ROLLBACK xid=None\n")
+        self.stream.flush()
+        self.mocker.replay()
+
+        connection = "CONNECTION"
+
+        self.tracer.connection_rollback(connection)
+
 
 class TimeoutTracerTestBase(TestHelper):
 
@@ -192,7 +242,9 @@ class TimeoutTracerTestBase(TestHelper):
 
         # Some data is kept in the connection, so we use a proxy to
         # allow things we don't care about here to happen.
-        class Connection(object): pass
+        class Connection(object):
+            pass
+
         self.connection = self.mocker.proxy(Connection())
 
     def tearDown(self):
@@ -330,6 +382,75 @@ class TimeoutTracerTest(TimeoutTracerTestBase):
         self.execute()
 
 
+class TimeoutTracerWithDBTest(TestHelper):
+
+    def setUp(self):
+        super(TimeoutTracerWithDBTest, self).setUp()
+        self.tracer = StuckInTimeTimeoutTracer(10)
+        install_tracer(self.tracer)
+        database = create_database(os.environ["STORM_POSTGRES_URI"])
+        self.connection = database.connect()
+
+    def tearDown(self):
+        super(TimeoutTracerWithDBTest, self).tearDown()
+        remove_tracer(self.tracer)
+        self.connection.close()
+
+    def is_supported(self):
+        return bool(os.environ.get("STORM_POSTGRES_URI"))
+
+    def test_timeout_set_on_beginning_of_new_transaction__commit(self):
+        """Check that we set the statement timeout before the first query of a
+        transaction regardless of the remaining time left by previous
+        transactions.
+
+        When we reuse a connection for a different transaction, the remaining
+        time of a previous transaction (which is stored in the connection)
+        could cause the first query in that transaction to run with no
+        timeout. This test makes sure that doesn't happen.
+        """
+        self.connection.execute('SELECT 1')
+        self.assertEqual([10], self.tracer.set_statement_timeout_calls)
+
+        self.connection.commit()
+
+        self.connection.execute('SELECT 1')
+        self.assertEqual([10, 10], self.tracer.set_statement_timeout_calls)
+
+    def test_timeout_set_on_beginning_of_new_transaction__rollback(self):
+        """Same as the test above, but here we rollback the first tx."""
+        self.connection.execute('SELECT 1')
+        self.assertEqual([10], self.tracer.set_statement_timeout_calls)
+
+        self.connection.rollback()
+
+        self.connection.execute('SELECT 1')
+        self.assertEqual([10, 10], self.tracer.set_statement_timeout_calls)
+
+
+class StuckInTimeTimeoutTracer(TimeoutTracer):
+
+    def __init__(self, fixed_remaining_time):
+        super(StuckInTimeTimeoutTracer, self).__init__()
+        self.set_statement_timeout_calls = []
+        self.fixed_remaining_time = fixed_remaining_time
+
+    def get_remaining_time(self):
+        return self.fixed_remaining_time
+
+    def set_statement_timeout(self, raw_cursor, remaining_time):
+        self.set_statement_timeout_calls.append(remaining_time)
+
+
+class StubConnection(Connection):
+
+    def __init__(self):
+        self._database = None
+        self._event = None
+        self._raw_connection = None
+        self.name = 'Foo'
+
+
 class BaseStatementTracerTest(TestCase):
 
     class LoggingBaseStatementTracer(BaseStatementTracer):
@@ -337,13 +458,6 @@ class BaseStatementTracerTest(TestCase):
             self.__dict__.setdefault('calls', []).append(
                 (connection, raw_cursor, statement))
 
-    class StubConnection(Connection):
-
-        def __init__(self):
-            self._database = None
-            self._event = None
-            self._raw_connection = None
-
     def test_no_params(self):
         """With no parameters the statement is passed through verbatim."""
         tracer = self.LoggingBaseStatementTracer()
@@ -352,7 +466,7 @@ class BaseStatementTracerTest(TestCase):
 
     def test_params_substituted_pyformat(self):
         tracer = self.LoggingBaseStatementTracer()
-        conn = self.StubConnection()
+        conn = StubConnection()
         conn.param_mark = '%s'
         var1 = MockVariable(u'VAR1')
         tracer.connection_raw_execute(
@@ -364,7 +478,7 @@ class BaseStatementTracerTest(TestCase):
     def test_params_substituted_single_string(self):
         """String parameters are formatted as a single quoted string."""
         tracer = self.LoggingBaseStatementTracer()
-        conn = self.StubConnection()
+        conn = StubConnection()
         var1 = MockVariable(u'VAR1')
         tracer.connection_raw_execute(
             conn, 'cursor', 'SELECT * FROM person where name = ?', [var1])
@@ -375,7 +489,7 @@ class BaseStatementTracerTest(TestCase):
     def test_qmark_percent_s_literal_preserved(self):
         """With ? parameters %s in the statement can be kept intact."""
         tracer = self.LoggingBaseStatementTracer()
-        conn = self.StubConnection()
+        conn = StubConnection()
         var1 = MockVariable(1)
         tracer.connection_raw_execute(
             conn, 'cursor',
@@ -388,7 +502,7 @@ class BaseStatementTracerTest(TestCase):
     def test_int_variable_as_int(self):
         """Int parameters are formatted as an int literal."""
         tracer = self.LoggingBaseStatementTracer()
-        conn = self.StubConnection()
+        conn = StubConnection()
         var1 = MockVariable(1)
         tracer.connection_raw_execute(
             conn, 'cursor', "SELECT * FROM person where id = ?", [var1])
@@ -399,7 +513,7 @@ class BaseStatementTracerTest(TestCase):
     def test_like_clause_preserved(self):
         """% operators in LIKE statements are preserved."""
         tracer = self.LoggingBaseStatementTracer()
-        conn = self.StubConnection()
+        conn = StubConnection()
         var1 = MockVariable(u'substring')
         tracer.connection_raw_execute(
             conn, 'cursor',
@@ -412,7 +526,7 @@ class BaseStatementTracerTest(TestCase):
 
     def test_unformattable_statements_are_handled(self):
         tracer = self.LoggingBaseStatementTracer()
-        conn = self.StubConnection()
+        conn = StubConnection()
         var1 = MockVariable(u'substring')
         tracer.connection_raw_execute(
             conn, 'cursor', "%s %s",
@@ -426,14 +540,14 @@ class BaseStatementTracerTest(TestCase):
 class TimelineTracerTest(TestHelper):
 
     def is_supported(self):
-        return timeline is not None
+        return has_timeline
 
     def factory(self):
         self.timeline = timeline.Timeline()
         return self.timeline
 
     def test_separate_tracers_own_state(self):
-        """"Check that multiple TimelineTracer's could be used at once."""
+        """Check that multiple TimelineTracer's could be used at once."""
         tracer1 = TimelineTracer(self.factory)
         tracer2 = TimelineTracer(self.factory)
         tracer1.threadinfo.action = 'foo'
@@ -457,8 +571,7 @@ class TimelineTracerTest(TestHelper):
 
     def test_finds_timeline_from_factory(self):
         factory_result = timeline.Timeline()
-        factory = lambda:factory_result
-        tracer = TimelineTracer(lambda:factory_result)
+        tracer = TimelineTracer(lambda: factory_result)
         tracer._expanded_raw_execute('conn', 'cursor', 'statement')
         self.assertEqual(1, len(factory_result.actions))
 
@@ -469,13 +582,6 @@ class TimelineTracerTest(TestHelper):
         self.assertEqual('statement', self.timeline.actions[-1].detail)
 
     def test_category_from_prefix_and_connection_name(self):
-        class StubConnection(Connection):
-
-            def __init__(self):
-                self._database = None
-                self._event = None
-                self._raw_connection = None
-                self.name = 'Foo'
         tracer = TimelineTracer(self.factory, prefix='bar-')
         tracer._expanded_raw_execute(StubConnection(), 'cursor', 'statement')
         self.assertEqual('bar-Foo', self.timeline.actions[-1].category)
@@ -491,3 +597,73 @@ class TimelineTracerTest(TestHelper):
         tracer = TimelineTracer(self.factory)
         tracer._expanded_raw_execute('conn', 'cursor', 'statement')
         self.assertEqual('SQL-<unknown>', self.timeline.actions[-1].category)
+
+
+class CaptureTracerTest(TestHelper, TestWithFixtures):
+
+    def is_supported(self):
+        return has_fixtures
+
+    def tearDown(self):
+        super(CaptureTracerTest, self).tearDown()
+        del _tracers[:]
+
+    def test_capture(self):
+        """
+        Using the L{CaptureTracer} fixture starts capturing queries and stops
+        removes the tracer upon cleanup.
+        """
+        tracer = self.useFixture(CaptureTracer())
+        self.assertEqual([tracer], get_tracers())
+        conn = StubConnection()
+        conn.param_mark = '%s'
+        var = MockVariable(u"var")
+        tracer.connection_raw_execute(conn, "cursor", "select %s", [var])
+        self.assertEqual(["select 'var'"], tracer.queries)
+
+        def check():
+            self.assertEqual([], get_tracers())
+
+        self.addCleanup(check)
+
+    def test_capture_as_context_manager(self):
+        """{CaptureTracer}s can be used as context managers."""
+        conn = StubConnection()
+        with CaptureTracer() as tracer:
+            self.assertEqual([tracer], get_tracers())
+            tracer.connection_raw_execute(conn, "cursor", "select", [])
+        self.assertEqual([], get_tracers())
+        self.assertEqual(["select"], tracer.queries)
+
+    def test_capture_multiple(self):
+        """L{CaptureTracer}s can be used as nested context managers."""
+
+        conn = StubConnection()
+
+        def trace(statement):
+            for tracer in get_tracers():
+                tracer.connection_raw_execute(conn, "cursor", statement, [])
+
+        with CaptureTracer() as tracer1:
+            trace("one")
+            with CaptureTracer() as tracer2:
+                trace("two")
+            trace("three")
+
+        self.assertEqual([], get_tracers())
+        self.assertEqual(["one", "two", "three"], tracer1.queries)
+        self.assertEqual(["two"], tracer2.queries)
+
+    def test_capture_with_exception(self):
+        """
+        L{CaptureTracer}s re-raise any error when used as context managers.
+        """
+        errors = []
+        try:
+            with CaptureTracer():
+                raise RuntimeError("boom")
+        except RuntimeError, error:
+            errors.append(error)
+        [error] = errors
+        self.assertEqual("boom", str(error))
+        self.assertEqual([], get_tracers())
